@@ -97,7 +97,6 @@ class TestBaseMCPAgent:
         assert agent.disallowed_tools == []
         assert agent.initial_screenshot is True
         assert agent.system_prompt is not None  # Default system prompt is set
-        assert agent.lifecycle_tools == []
 
     def test_init_with_params(self):
         """Test initialization with custom parameters."""
@@ -108,7 +107,6 @@ class TestBaseMCPAgent:
             disallowed_tools=["bad_tool"],
             initial_screenshot=True,
             system_prompt="Custom prompt",
-            lifecycle_tools=["custom_setup", "custom_eval"],
         )
 
         assert agent.mcp_client == client
@@ -116,7 +114,6 @@ class TestBaseMCPAgent:
         assert agent.disallowed_tools == ["bad_tool"]
         assert agent.initial_screenshot is True
         assert agent.system_prompt == "Custom prompt"
-        assert agent.lifecycle_tools == ["custom_setup", "custom_eval"]
 
     @pytest.mark.asyncio
     async def test_init_no_client_no_task(self):
@@ -329,9 +326,6 @@ class TestBaseMCPAgent:
         """Test getting tool schemas."""
         agent = MockMCPAgent()
 
-        # Add setup to lifecycle tools to test filtering
-        agent.lifecycle_tools = ["setup"]
-
         agent._available_tools = [
             types.Tool(name="tool1", description="Tool 1", inputSchema={"type": "object"}),
             types.Tool(name="setup", description="Setup", inputSchema={"type": "object"}),
@@ -601,7 +595,7 @@ class TestMCPAgentExtended:
         agent = MockAgentExtended(mcp_client=mock_client, allowed_tools=["tool1", "tool3"])
         await agent.initialize("test")
 
-        available_names = [tool.name for tool in agent._available_tools]
+        available_names = [tool.name for tool in agent.get_available_tools()]
         assert "tool1" in available_names
         assert "tool3" in available_names
         assert "tool2" not in available_names
@@ -620,7 +614,7 @@ class TestMCPAgentExtended:
         agent = MockAgentExtended(mcp_client=mock_client, disallowed_tools=["tool2"])
         await agent.initialize("test")
 
-        available_names = [tool.name for tool in agent._available_tools]
+        available_names = [tool.name for tool in agent.get_available_tools()]
         assert "tool1" in available_names
         assert "tool3" in available_names
         assert "tool2" not in available_names
@@ -631,7 +625,6 @@ class TestMCPAgentExtended:
         # Lifecycle tools are specified by name, not as objects
         agent = MockAgentExtended(
             mcp_client=mock_client,
-            lifecycle_tools=["screenshot"],  # Use tool name
             responses=[{"role": "assistant", "content": "Done", "tool_calls": []}],
         )
 
